package Encode::Registry;

use strict;
use XML::Parser::Expat;
use IO::File;
use File::Spec;

our %platforms;
our %mappings;
our %aliases;
our $base_class = 'Encode';
our $registry;

sub import
{
    my ($class, %opts) = @_;
    my ($regfile);

    $regfile = $opts{'file'} || $class->find_registry;
    $class->read_file($regfile) if ($regfile);
}

sub find_registry
{
    my (@path, $p, $file);
    
    if ($^O eq 'MSWin32')
    {
        require Win32::TieRegistry;
        Win32::TieRegistry->import(Delimiter => '/');

        $file = $registry->{"LMachine/Software/Mappings/Registry//File"};
        return $file if $file;
    }

    @path = split(/;\s*/, $ENV{'MAPPINGPATH'});
    foreach $p (@path)
    {
        $file = File::Spec->catfile($p, 'REGISTRY.XML');
        return $file if (-f $file);
    }
    return undef;
}
        
sub read_file
{
    my ($class, $fname) = @_;
    my ($xml) = XML::Parser::Expat->new();
    my ($curr_plat, $curr_map);

    $xml->setHandlers('Start' => sub {
        my ($xml, $tag, %attrs) = @_;
        
        if ($tag eq 'platform')
        {
            $curr_plat = {};
            $platforms{$attrs{'name'}}{$attrs{'factory'}} = $curr_plat;
        }
        elsif ($tag eq 'implement')
        {
            my ($id) = $attrs{'id'};
            
            push (@{$curr_plat->{$attrs{'id'}}}, [$attrs{'use'}, $attrs{'priority'}]);
        }
        elsif ($tag eq 'map')
        {
            if (defined $attrs{'see'})
            { $aliases{$attrs{'name'}} = $attrs{'see'}; }
            else
            {
                $curr_map = {};
                $mappings{$attrs{'name'}} = $curr_map;
            }
        }
        elsif ($tag eq 'src')
        {
            push (@{$curr_map->{$attrs{'type'}}}, [$attrs{'system'}, $attrs{'public'}]);
        }
    }, 'ExternEnt' => sub {
        my ($xml, $base, $sys, $pub) = @_;
        my ($fname) = File::Spec->rel2abs($sys, $base);
        my ($fh) = IO::File->new("> $fname");
        return $fh;
    });

    $xml->parsefile($fname);
}


sub write_file
{
    my ($class, $fname) = @_;
    my ($fh) = IO::File->new("> $fname") || return undef;
    my ($p, $f, $i, $j);

    $fh->print("<?xml version='1.0'?>\n");
    $fh->print("<!-- Autogenerated by Perl, Encode::Registry at " . scalar localtime() . "-->\n");
    $fh->print("<registry>\n\n");

    $fh->print("<implementations>\n\n");
    foreach $p (sort keys %platforms)
    {
        foreach $f (sort keys %{$platforms{$p}})
        {
            $fh->print("    <platform name=\"$p\" factory=\"$f\"/>\n");
            foreach $i (sort keys %{$platforms{$p}{$f}})
            {
                foreach $j (@{$platforms{$p}{$f}{$i}})
                {
                    $fh->print("        <implement id=\"$i\" use=\"$j->[0]\"");
                    $fh->print(" priority=\"$j->[1]\"") if ($j->[1]);
                    $fh->print("/>\n");
                }
            }
            $fh->print("    </platform>\n\n");
        }
    }
    $fh->print("</implementations>\n\n");

    $fh->print("<mappings>\n\n");
    foreach $p (sort (keys %mappings, keys %aliases))
    {
        if (defined $aliases{$p})
        {
            $fh->print("    <map name=\"$p\" see=\"$aliases{$p}\"/>\n\n");
        } else
        {
            $fh->print("    <map name=\"$p\">\n");
            foreach $f (sort keys %{$mappings{$p}})
            {
                foreach $i (@{$mappings{$p}{$f}})
                {
                    $fh->print("        <src type=\"$f\" system=\"$i->[0]\" public=\"$i->[1]\"/>\n");
                }
            }
            $fh->print("    </map>\n\n");
        }
    }
    $fh->print("</mappings>\n\n");

    $fh->print("</registry>\n");
    $fh->close;
    return 1;
}


sub add_handler
{
    my ($class, $type, $handler, $priority) = @_;

    push (@{$platforms{'perl'}{$base_class}{$type}}, [$handler, $priority]);
}


sub get_encoding
{
    my ($class, $name) = @_;
    my ($t, $p, $h, %temp, $res);

    while (defined $aliases{$name})
    { $name = $aliases{$name}; }

    foreach $t (keys %{$mappings{$name}})
    {
        foreach $p (@{$mappings{$name}{$t}})
        { last if ($p->[0]); }
        next unless ($p->[0]);

        $h = (sort {$b->[1] <=> $a->[1]} @{$platforms{'perl'}{$base_class}{$t}})[0];
        $temp{$h->[0]} = [$p->[0], $h->[1]];
    }

    $h = (sort {$temp{$b}[1] <=> $temp{$a}[1]} keys %temp)[0];

    $res = {$h}->new($temp{$h}[0]);
    return $res;
}


sub add_encoding
{
    my ($class, $name, $type, $lfile, $pfile) = @_;

    push (@{$mappings{$name}{$type}}, [$lfile, $pfile]);
}

sub add_alias
{
    my ($class, $name, $alias) = @_;

    $aliases{$name} = $alias;
}

__END__

=head1 NAME

Encode::Registry - Registry file handling for character encodings

=head1 SYNOPSIS

    use Encode::Registry;

    $enc = Encode::Registry->get_encoding($enc_name);

=head1 DESCRIPTION

The C<Encode> module provides a Perl interface to the cross-architecture
registry of character encoding information. This registry takes the form of an
XML file containing information about encoding files, their types and handlers
for different file types on different platforms (or programming environments).

=head2 Registry File

The XML file has the following structure:

    <?xml version='1.0'?>
    <registry>

    <implementations>
        <platform name="perl" factory="Encode">
            <implement id="www.tcl.org/encoding/enc" use="Encoding::Tcl" priority="3"/>
            <implement id="www.perl.com/CPAN/Unicode/Map" use="Unicode::Map" priority="13"/>
        </platform>
        <platform name="COM" factory="7712d4d4-fd95-4e5e-900e-346dbca868ae">
            <implement id="www.unicode.org/TR22"
                    use="5f95df81-a2e7-495f-ae81-93be149ac21d"/>
        </platform>
    </implementations>

    <mappings>
        <map name='CP1252'>
            <src type='www.unicode.org/MAPPINGS/text'
                    public='ftp://ftp.unicode.org/Public/MAPPINGS/VENDORS/MICSFT/WINDOWS/CP1252.TXT'
                    system='C:/Unicode/Mappings/Vendors/Micsft/Windows/cp1252.txt'/>
            <src type='www.perl.com/CPAN/Unicode/Map'
                    system='C:/Progs/Perl/site/lib/Unicode/Map/MS/WIN/CP1252.map'/>
            <src type='www.tcl.org/encoding/enc'
                    system='C:/Progs/Perl/lib/Encode/cp1252.enc'/>
        </map>
        <map name="cp-1252" see="CP1252"/>
    </mappings>
    
    </registry>

The XML file consists of two parts. The first part is a set of platform elements
which describe how a particular platform and environment implement different
encoding description types. Thus, for the C<perl> example, the particular C<perl>
environment we are concerned with is the C<Encode> module. For this module, the
TCL encoding (C<.enc>) files can be processed using C<Encoding::Tcl> and the
Map files (C<.map>) can be used with C<Unicode::Map>). The C<priority> attributes
are used to indicate which implementation is more preferable. The higher the value,
the more preferable it is to use that implementation.

The second part of the file consists of mapping descriptions. A mapping description
may be an alias, in which case the C<map> element is empty, but there is a
C<see> attribute to cross-refer to another c<map> element. A normal mapping
description consists of a number of C<src> elements. Each C<src> element
corresponds to a file somewhere. Usually this would be a file on the system, in
which case the C<src> element will have a C<system> attribute to reference it.
The C<src> element may also have a C<public> attribute to reference the same
file via a URL. The C<type> attribute indicates which type of file the file is
and is used by the implementation section to decide how such a file should be
handled.

The DTD for this XML description is presented later.

=head2 Encoding Types

The only requirement is that all files of the same file structure have the same type
identifier. This identifier is best based on some URL, even if that URL doesn't itself
actually exist. The following is a list of file formats I have given type identifiers
to subject to an industry wide registry being set up.

    oss.software.ibm.com/icu            .icu files from ICU
    www.dkuug.dk/i18n/charmaps          Character maps from ISO
    www.perl.com/CPAN/Unicode/Map       Binary files from Unicode::Map module
    www.perl.com/CPAN/Unicode/Map8      Binary files from Unicode::Map8
    www.tcl.org/encodings/enc           .enc format from TCL
    www.unicode.org/MAPPINGS/text       Normal Unicode Text files mapping description
    www.unicode.org/MAPPINGS/reverse    As above with the columns reversed
    www.unicode.org/MAPPINGS/jis        As above using columns 2 & 3 (for JIS encodings)

=head2 Locating the Registry file

Having a central registry for encodings is very useful, assuming you can find it.
But tracking down such a registry given that it may be shared by many programming
environments, is not easy.

On the Windows platform, this module will look for the registry file in the system
registry under the following key

    HKEY_LOCAL_MACHINE\SOFTWARE\Mapping\Registry\File

which is held as a string.

Failing this, and on other systems, the module will examine the C<MAPPINGPATH>
environment variable and look in each directory listed for a file C<REGISTRY.xml>.
Directories in the environment variable are separated by a ;

=head1 MODULE

To use the module, simply C<use> it.

    use Encode::Registry

This will cause the C<import> subroutine to run which will search for a REGISTRY.xml
and read it. If nothing is found, then nothing will be read.

To load the highest priority implementation of an encoding, use:

    $enc = Encode::Registry->get_encoding($enc_name);

=head2 Methods

=head3 Encode::Registry->find_registry

Returns an absolute file name reference to the registry file, if it can find it.

=head3 Encode::Registry->read_file($fname)

Reads and processes the given XML file adding the information to the internal
data structures of this class.

=head3 Encode::Registry->write_file($fname)

Writes a registry file to $fname containing all the information in the internal
data structures.

=head3 Encode::Registry->add_handler($type, $handler, $priority)

Adds the given type to be handled by the handler module when using Perl. The
priority is also set this way.

=head3 Encode::Registry->get_encoding($name)

Returns an object capable of processing data of the given encoding. The object
is created by calling

    $handler->new($system_file)

Where C<$handler> is the associated handler for the given file type of
C<$system_file>. The C<$handler> with the highest priority is chosen.

=head3 Encode::Registry->add_encoding($name, $type, $local_file, $public_url)

Adds a source relationship for C<$local_file> of type C<$type> with C<$public_url>
against the mapping C<$name>.

=head3 Encode::Registry->add_alias($name, $alias)

Adds an alias C<$name> which is an alias of C<$alias>

=head1 DTD

The DTD for the registry XML file is:

    <?xml version="1.0" encoding="UTF8"?>
    <!DOCTYPE registry [

    <!ELEMENT registry (implementations, mappings)>

    <!ELEMENT implmentations (platform+)>

    <!ELEMENT platform (implement+)>
    <!ATTLIST platform
        name    CDATA   #REQUIRED
        factory CDATA   #REQUIRED
    >

    <!ELEMENT implement EMPTY>
    <!ATTLIST implement
        id      CDATA   #REQUIRED
        use     CDATA   #REQUIRED
        priority CDATA  #IMPLIED
    >

    <!ELEMENT mappings (mapping+)>

    <!ELEMENT map (src)*>
    <!ATTLIST map
        name    ID      #REQUIRED
        see     IDREF   #IMPLIED
    >

    <!ELEMENT src EMPTY>
    <!ATTLIST src
        type    CDATA
        system  CDATA
        public  CDATA
    >

    ]>

